# -*- coding: utf-8 -*-
# :Project:   metapensiero.sphinx.autodoc_sa -- Fixtures used by the test suite
# :Created:   lun 22 feb 2016 14:18:21 CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2016, 2017 Lele Gaifax
#

from __future__ import unicode_literals

from io import open
from locale import getpreferredencoding
from os.path import join
import shutil
import subprocess
import tempfile
import unittest


USER_ENCODING = getpreferredencoding() or "UTF-8"

DEFAULT_SPHINX_CONF = """\
import os, sys
sys.path.insert(0, os.path.abspath(os.path.dirname(__file__)))
extensions = ['sphinx.ext.autodoc', 'metapensiero.sphinx.autodoc_sa']
source_suffix = '.txt'
master_doc = 'index'
"""

DEFAULT_INDEX_TXT = """\
Autodoc SA test
===============

.. automodule:: test
   :members:
"""


DEFAULT_TEST_PY = """\
from sqlalchemy import Column, Date, Integer, MetaData, String, Table, func, orm, select

metadata = MetaData()

persons = Table('persons', metadata,
                Column('id', Integer, primary_key=True),
                Column('firstname', String),
                Column('lastname', String),
                Column('birthdate', Date),
                )

class Person(object):
    def __init__(self, firstname, lastname, birthdate):
        self.firstname = firstname
        self.lastname = lastname
        self.birthdate = birthdate

orm.mapper(Person, persons)

FOO = None
'Bar'
"""


class TestSphinx(object):
    SPHINX_BUILD = 'sphinx-build'
    SPHINX_BUILD_OPTS = ('-b', 'html', '-q', '-d', '_build/doctrees', '.', '_build/html')

    def __init__(self, db_opts=None):
        self.db_opts = db_opts
        self.directory = tempfile.mkdtemp()

    def remove(self):
        shutil.rmtree(self.directory)

    def build(self, contents):
        from textwrap import dedent

        for filename, content in contents.items():
            with open(join(self.directory, filename), 'w', encoding='utf-8') as f:
                f.write(dedent(content))

        cmd = [self.SPHINX_BUILD]
        cmd.extend(self.SPHINX_BUILD_OPTS)
        try:
            output = subprocess.check_output(cmd, cwd=self.directory, stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as e:
            self.build_error = e.output.decode(USER_ENCODING)
        else:
            output = output.decode(USER_ENCODING)
            self.build_error = output or None
        return self.build_error

    def get_artifact(self, name):
        fname = join(self.directory, '_build', 'html', name)
        with open(fname, encoding='utf-8') as f:
            return f.read()


class BaseTestCase(unittest.TestCase):
    SPHINX_CONF = DEFAULT_SPHINX_CONF
    "The configuration for the Sphinx environment"

    INDEX_TXT = DEFAULT_INDEX_TXT
    "The entry point of the documentation"

    TEST_PY = DEFAULT_TEST_PY
    "The test module"

    OTHER_FILES = ()
    "A possible sequence of ``(filename, content)`` tuples"

    @classmethod
    def contents(cls):
        yield 'conf.py', cls.SPHINX_CONF
        yield 'index.txt', cls.INDEX_TXT
        test_py = getattr(cls, 'TEST_PY', None)
        if test_py is not None:
            yield 'test.py', test_py
        for fname, content in cls.OTHER_FILES:
            yield fname, content

    @classmethod
    def setUpClass(cls):
        cls.sphinx = TestSphinx()
        cls.sphinx.build({filename: content for filename, content in cls.contents()})

    @classmethod
    def tearDownClass(cls):
        cls.sphinx.remove()

    @property
    def build_error(self):
        return self.sphinx.build_error

    def setUp(self):
        self.assertIsNone(self.build_error)

    def build(self, contents):
        return self.sphinx.build(contents)

    def get_artifact(self, name):
        return self.sphinx.get_artifact(name)

    @property
    def index_html(self):
        return self.get_artifact('index.html')



# -*- coding: utf-8 -*-
# :Project:   metapensiero.sphinx.autodoc_sa -- Test with autoclass
# :Created:   mar 17 gen 2017 09:08:45 CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2017 Lele Gaifax
#

from __future__ import unicode_literals

import fixtures


SELECT = """

Q = select([persons])

class Class:
    'This is a class'

    SELECT = Q
    'This is a select statement'

"""

class TestClass(fixtures.BaseTestCase):
    TEST_PY = fixtures.BaseTestCase.TEST_PY + SELECT
    INDEX_TXT = """\
Autodoc SA test
===============

.. autoclass:: test.Class
   :members:

"""

    def test(self):
        index_html = self.index_html
        assert 'id="test.Class.SELECT"' in index_html
        assert 'This is a select' in index_html
        assert '<span class="k">SELECT</span>' in index_html



# -*- coding: utf-8 -*-
# :Project:   metapensiero.sphinx.autodoc_sa -- Test for the ORM side
# :Created:   mar 17 gen 2017 00:25:10 CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2017 Lele Gaifax
#

import fixtures


SELECT = """

Q = orm.Query(Person)
'This is an ORM query'

"""

class TestSelect(fixtures.BaseTestCase):
    TEST_PY = fixtures.BaseTestCase.TEST_PY + SELECT

    def test(self):
        index_html = self.index_html
        assert 'id="test.Q"' in index_html
        assert 'This is an ORM query' in index_html
        assert '<span class="k">SELECT</span>' in index_html



# -*- coding: utf-8 -*-
# :Project:   metapensiero.sphinx.autodoc_sa -- Pretty print canned SQLAlchemy statements
# :Created:   mar 11 ago 2015 11:44:44 CEST
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2015, 2016, 2017 Lele Gaifax
#

from sphinx.ext import autodoc
from sqlalchemy.sql import ClauseElement
from sqlalchemy.orm import Query


try:
    from sqlparse import format as sqlparse_format
except ImportError:
    sqlparse_prettifier = None
else:
    def sqlparse_prettifier(sql):
        return sqlparse_format(sql, reindent=True)

try:
    from pg_query import prettify
except ImportError:
    pg_query_prettifier = None
else:
    def pg_query_prettifier(sql):
        return prettify(sql)


dialect = None

def select_sa_dialect(app):
    global dialect
    classname = app.config.autodoc_sa_dialect
    if classname:
        if isinstance(classname, str):
            modulename, classname = classname.rsplit('.', 1)
            module = __import__(modulename, fromlist=[classname])
            dialect = getattr(module, classname)()
        else:
            dialect = classname
    else:
        dialect = None


future_param = object()


def format_value(value):
    from uuid import UUID

    if isinstance(value, UUID):
        value = str(value)

    result = repr(value)

    if len(result) > 50:
        result = result[:20] + ' … ' + result[-20:]

    return result


def format_numeric_param(position, prefix, args, match):
    key = int(match.group(1))
    value = args[key-1]
    if value is future_param:
        return '%s%s' % (prefix, next(position))
    else:
        return format_value(value)


def format_named_param(args, match):
    key = match.group(1)
    value = args[key]
    if value is future_param:
        return ':' + key
    else:
        return format_value(value)


def interpolate_sa_params(compiled_stmt):
    from functools import partial
    from itertools import count
    from re import escape, sub

    sql = compiled_stmt.string

    # This is a dirty hack, at best: SA injects bindparams by its own, to pass literal values
    # to the DB; here we try to "simplify" the query, replacing them with their known values,
    # leaving "real" bindparams (that is, the ones that developers coded in the query using
    # bindparam() explicitly. That is bad and errorprone, but after all we just want to pretty
    # print the statement for documentation purposes and you know, nobody RTFM anyway...
    #
    # Things may change in the future, see
    # https://bitbucket.org/zzzeek/sqlalchemy/issues/4024/mention-the-advanced-technique-to-have-raw

    params = {}
    for bp, bn in compiled_stmt.bind_names.items():
        if bp.key.startswith('%('):
            params[bn] = bp.value
        else:
            params[bn] = future_param

    if params:
        pstyle = dialect.paramstyle
        if pstyle == 'numeric':
            args = tuple(params[p] for p in compiled_stmt.positiontup)
            ptemplate = compiled_stmt.sql_compiler.bindtemplate
            format = partial(format_numeric_param, count(1), ptemplate[0])
            pregexp = r'%s(\d+)' % escape(ptemplate[0])
        elif pstyle == 'qmark':
            raise NotImplementedError("Paramstyle %r not implemented, sorry" % pstyle)
        elif pstyle == 'pyformat':
            args = params
            format = format_named_param
            pregexp = r'%%\((%s)\)s' % '|'.join(args.keys())
        elif pstyle == 'named':
            args = params
            format = format_named_param
            pregexp = r':(%s)' % '|'.join(args.keys())

        sql = sub(pregexp, lambda m: format(args, m), sql)

    return sql


def stringify(env, stmt):
    how = env.config.autodoc_sa_prettifier
    if how == 'sqlparse':
        prettifier = sqlparse_prettifier
    elif how == 'pg_query':
        prettifier = pg_query_prettifier
    else:
        prettifier = None

    if isinstance(stmt, Query):
        stmt = stmt.statement

    try:
        sql = interpolate_sa_params(stmt.compile(dialect=dialect))
    except Exception as e:
        env.warn(env.docname,
                 "Parameters interpolation raised an error, keeping bogus statement: %s"
                 % e)
        sql = str(stmt)

    if prettifier is not None:
        try:
            sql = prettifier(sql)
        except Exception as e:
            env.warn(env.docname,
                     "SQL prettification raised an error, keeping the original: %s"
                     % e)
            pass

    return sql


class AttributeDocumenter(autodoc.AttributeDocumenter):
    """
    Customized AttributeDocumenter that knows about SA Select
    """

    def add_directive_header(self, sig):
        if not isinstance(self.object, (ClauseElement, Query)):
            autodoc.AttributeDocumenter.add_directive_header(self, sig)
        else:
            autodoc.ClassLevelDocumenter.add_directive_header(self, sig)

    def add_content(self, more_content, no_docstring=False):
        autodoc.AttributeDocumenter.add_content(self, more_content, no_docstring)
        if isinstance(self.object, (ClauseElement, Query)):
            sql = stringify(self.env, self.object)
            self.add_line("", "")
            lang = self.env.config.autodoc_sa_pygments_lang
            self.add_line(".. code-block:: %s" % lang, "")
            self.add_line("", "")
            for line in sql.splitlines():
                self.add_line("   " + line, "")


class DataDocumenter(autodoc.DataDocumenter):
    """
    Customized DataDocumenter that knows about SA Select
    """

    def add_directive_header(self, sig):
        if not isinstance(self.object, (ClauseElement, Query)):
            autodoc.DataDocumenter.add_directive_header(self, sig)
        else:
            autodoc.ModuleLevelDocumenter.add_directive_header(self, sig)

    def add_content(self, more_content, no_docstring=False):
        autodoc.DataDocumenter.add_content(self, more_content, no_docstring)
        if isinstance(self.object, (ClauseElement, Query)):
            sql = stringify(self.env, self.object)
            self.add_line("", "")
            lang = self.env.config.autodoc_sa_pygments_lang
            self.add_line(".. code-block:: %s" % lang, "")
            self.add_line("", "")
            for line in sql.splitlines():
                self.add_line("   " + line, "")


def setup(app):
    "Setup the Sphinx environment."

    from sphinx.config import ENUM

    autodoc.add_documenter(AttributeDocumenter)
    autodoc.add_documenter(DataDocumenter)
    app.add_config_value('autodoc_sa_dialect', None, True)
    app.add_config_value('autodoc_sa_prettifier', 'sqlparse', True,
                         ENUM('sqlparse', 'pg_query'))
    app.add_config_value('autodoc_sa_pygments_lang', 'sql', True)
    app.connect('builder-inited', select_sa_dialect)



# -*- coding: utf-8 -*-
# :Project:   metapensiero.sphinx.autodoc_sa -- Test pg_query prettifier
# :Created:   lun 07 ago 2017 16:44:22 CEST
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2017 Lele Gaifax
#

import sys

import pytest

import fixtures


SELECT = """

p = persons.alias('p')
p2 = persons.alias('p2')
Q = select([p.c.firstname, p.c.lastname,
            select([p2.c.lastname])\
            .where(p2.c.id == p.c.id).as_scalar()]) \
    .where(func.check_name(p.c.firstname))
'This is a silly select statement'

"""

@pytest.mark.skipif(sys.version_info < (3,6), reason="pg_query requires at least Py3.6")
class Test_PG_Query_Prettify(fixtures.BaseTestCase):
    TEST_PY = fixtures.BaseTestCase.TEST_PY + SELECT
    SPHINX_CONF = fixtures.BaseTestCase.SPHINX_CONF + \
                  '\nautodoc_sa_dialect = "sqlalchemy.dialects.postgresql.dialect"' + \
                  '\nautodoc_sa_prettifier = "pg_query"' + \
                  '\nautodoc_sa_pygments_lang = "postgresql"'

    def test(self):
        index_html = self.index_html
        assert '     <span class="p">,</span> <span class="n">p</span>' in index_html
        assert '     <span class="p">,</span> <span class="p">(</span><span class="k">SELECT</span>' in index_html
        assert '        <span class="k">FROM</span> <span class="n">persons</span>' in index_html
